;;; $Id: commands.scm,v 1.36 2003/02/22 17:13:45 queinnec Exp $
;;; Copyright (c) 1990-2001 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; This file defines the commands offered by LiSP2TeX
;;; These commands are:
;;;    FromFile                 ( file keys... )
;;;    Print                    Sexpression
;;;    PrettyPrint              Sexpression
;;;    PrettyFromFile           ( file keys... )
;;;    Eval                     Sexpression
;;;    ShowEvaluationOf		Sexpression
;;;    FromCFile                ( file strings... )
;;;    InsertFile               ( file ... )
;;;    LoadFile                 ( file ... )
;;;    ExtractFromFile          ( file keys... )

;;; Any command <C> is associated to two functions: run-<C> and
;;; command-<C>. Whenever a directive is found in the input stream,
;;; the function run-<C> is invoked to read some parameters and apply
;;; command-<C> on them.

;;; Directives that prefix a list decompose that list into options and
;;; arguments. Arguments are specified as above. Options are gathered
;;; in a list whose first term is a keyword. Example, you may write
;;;    \FromFile("foo.scm" foo) 
;;; or \FromFile((snippetenv: foo snippetIndent: 3) "foo.scm" foo)
;;; Keywords are followed by exactly one value. Keywords are symbols
;;; suffixed with a colon.
;;;      ATTENTION: \FromFile( () "" ...) is not \FromFile("" ...)

(define (get-options e)
  (if (pair? e)
      (cond ((null? (car e)) 
             *last-options* )
            ((and (pair? (car e))
                  (looks-like-keyword? (caar e)) )
             (set! *last-options* (car e))
             (car e) )
            (else 
             (set! *last-options* '())
             '()) )
      '() ) )

;;; record last options so it can be reused with ()
(define *last-options* '())

(define (get-arguments e)
  (if (pair? e)
      (if (or (null? (car e))
              (and (pair? (car e))
                   (looks-like-keyword? (caar e)) ) )
          (cdr e)
          e )
      e ) )

;;; retrieve in a list of options the value associate with the keyword name:
;;; `name' should be a string not ending with colon.

(define (get-option-value options name)
  (let ((kw (build-keyword name)))
    (define (gov options)
      (if (pair? options)
          (if (pair? (cdr options))
              (if (eq? (car options) kw)
                  (cadr options)
                  (gov (cddr options)) )
              (command-error *get-options-error* options) )
          #f ) )
    (gov options) ) )

(define (build-keyword name)
  (cond 
   ((looks-like-keyword? name) name)
   ((string? name) (string->symbol (string-append name ":")))
   ((symbol? name) (string->symbol (string-append (symbol->string name) ":")))
   (else (command-error *get-option-value-error* name)) ) )

(define (looks-like-keyword? e)
  (and (symbol? e)
       (char=? #\: (let ((s (symbol->string e)))
                     (string-ref s (- (string-length s) 1)) )) ) )

(define *get-options-error* 
  "~%LiSP2TeX: Incorrect options ~A~%" )
(set! *get-options-error*  *get-options-error*)
(define *get-option-value-error* 
  "~%LiSP2TeX: Incorrect option keyword ~A~%" )
(set! *get-option-value-error* *get-option-value-error*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Print
;;; Read and print a Sexpression. By default (see port-XXX.scm) try to
;;; respect input case.

(define (run-Print stream)
  (command-Print (numbering-read stream)) )

(define (command-Print exp)
  (pp-format stdout-port *lisp-excerpt* exp) )

(define *lisp-excerpt* "{\\snip[scheme]{~A}}")
(set! *lisp-excerpt* *lisp-excerpt*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; PrettyPrint
;;; Read and pretty-print a Sexpression. By default (see port-XXX.scm)
;;; try to respect input case.

(define (run-PrettyPrint stream)
  (command-PrettyPrint (numbering-read stream)) )

(define (command-PrettyPrint exp)
  (pp-greekify exp stdout-port) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Insert a complete file
;;; Insert a complete file duly surrounded with directives.

(define (run-InsertFile stream)
  (let ((e (numbering-read stream)))
    (let ((options   (get-options e))
          (arguments (get-arguments e)) )
      (if (and (pair? arguments) (every? string? arguments))
          (command-InsertFile options arguments)
          (command-error *insertfile-error1* e) ) ) ) )

(define *insertfile-error1*
  "~%LiSP2TeX (InsertFile) error: Illegal filename ~A~%" )
(set! *insertfile-error1* *insertfile-error1*)
(define *insertfile-error2*
  "~%LiSP2TeX (InsertFile) error: No such filename ~A~%" )
(set! *insertfile-error2* *insertfile-error2*)

(define (command-InsertFile options filenames)
  (let ((snippetenv (get-option-value options "snippetenv")))
    (define (sai filename)
      (search-and-insert snippetenv filename) )
    (for-each sai filenames) ) )

(define (search-and-insert snippetenv filename)
  (command-echo "[Inserting ~U within (snippetenv: ~A) ..." 
                filename snippetenv )
  (let ((file (search-file filename *path*)))
    (if file
        (let ((string (file->string file)))
          (pp-format stdout-port *insertfile-format* 
                     filename 
                     (or snippetenv
                         (snippet-environment file) )
                     string )
          (command-echo "]") )
        (command-error *insertfile-error2* filename) ) ) )

(define *insertfile-format* 
  "% Inserted File ~A~%\\snippet[~U]~%~U\\endsnippet" )
(set! *insertfile-format* *insertfile-format*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; LoadFile
;;; Read the name of a scheme file and load it. The difference with
;;; \Eval (load "file") is that the file is looked for in the current
;;; path. 

(define (run-LoadFile stream)
  (let ((e (numbering-read stream)))
    ;; ignore options
    (let ((e (get-arguments e)))
      (if (and (pair? e) (every? string? e))
          (command-LoadFile e)
          (command-error *loadfile-error1* e) ) ) ) )

(define (every? p e)
  (if (pair? e)
      (and (p (car e))
           (every? p (cdr e)) )
      #t ) )

(define *loadfile-error1*
  "~%LiSP2TeX (Loadfile) error: Illegal filename within ~A~%" )
(set! *loadfile-error1* *loadfile-error1*)

(define (command-LoadFile filenames)
  (for-each search-and-load filenames) )

(define (search-and-load filename)
  (let ((tmp-file (get-tmp-file)))
    (command-echo "[Loading ~U (output into ~U) ..." filename tmp-file)
    (let ((file (search-file filename *path*)))
      (if file
          (begin 
            (with-output-to-file tmp-file
              (lambda () (load-parameter-file file)) )
            (command-echo "]")
            (delete-file tmp-file) )
          (command-error *loadfile-error2* filename) ) ) ) )

(define *loadfile-error2*
  "~%LiSP2TeX (Loadfile) error: No such filename ~A~%" )
(set! *loadfile-error2* *loadfile-error2*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; PrettyPrintFromFile
;;; Read from a file and pretty-print a Sexpression. As for FromFile
;;; there is a concept of current file (held in the global variable
;;; *the-current-scanned-file*).

(define (run-PrettyFromFile stream)
  (let ((e (numbering-read stream)))
    ;; ignore options
    (let ((e (get-arguments e)))
      (if (pair? e)
          (if (string? (car e))
              (command-PrettyFromFile (car e) (cdr e))
              (command-error *fromfile-error1* "PrettyFromFile" e) )
          (command-error *fromfile-error2* "PrettyFromFile" e) ) ) ) )

(define (command-PrettyFromFile filename keys)
  (let ((sf (determine-scanned-file "PrettyFromFile" filename scan-file)))
    (if sf
        (let ((ses (mapcan (lambda (key) (find-scanned-expression key sf))
                           keys )) ) 
          (for-each (lambda (se)
                      (if (eq? se '***MISSING-EXPRESSION***)
                          (pp-greekify (symbol->string se) stdout-port)
                          (pp-greekify (scanned-expression-original se) 
                                       stdout-port ) ) )
                    ses ) ) ) ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ShowEvaluationOf
;;; Read, print, evaluate an Sexpression then print its result.  Since
;;; there is an evaluation performed by the underlying Scheme, the
;;; expression is read with the preferred case sensibility of the
;;; underlying Scheme.

(define (run-ShowEvaluationOf stream)
  ;; save the current settings of the reader
  (let* ((ri *respect-input*)
         (ui *uppercase-input*)
         (li *lowercase-input*)
         (start-line (numbering-stream-line stream)) )
    ;; sets the reader to the way eval prefers
    (set! *respect-input*   *evaluator-preferred-respect-input*)
    (set! *uppercase-input* *evaluator-preferred-uppercase-input*)
    (set! *lowercase-input* *evaluator-preferred-lowercase-input*)
    (let* ((exp (numbering-read stream))
           (end-line (numbering-stream-line stream)) )
      (pp-format stdout-port *showevaluationof-input* exp)
      ;; restaure the reader
      (set! *respect-input*   ri)
      (set! *uppercase-input* ui)
      (set! *lowercase-input* li)
      ;; and only now evaluates the read expression
      (let ((result (LiSP2TeX-eval exp)))
        (if (and (pair? result)
                 (eq? (car result) '***EVALUATION-ERROR***) )
            (command-error *showevaluationof-error* 
                           start-line
                           end-line
                           exp
                           (cdr result) ) )
        ;; ignore or print:
        (if (ignore? result)
            (pp-format stdout-port
                       *showevaluationof-ignore-output*
                       result )
            (pp-format stdout-port
                       *showevaluationof-output* 
                       result ) ) ) ) ) )

;;; In Gambit, define #<void> to be ignored:

(define values-to-ignore 
  (list (if #f 'ignore)) )
(set! values-to-ignore values-to-ignore)

(define (ignore? v)
  (memq v values-to-ignore) )

(define *showevaluationof-input*    "\\snip[scheme]{~S} ")
(set! *showevaluationof-input* *showevaluationof-input*)
(define *showevaluationof-output*   " gives \\snip[scheme]{~S}")
(set! *showevaluationof-output* *showevaluationof-output*)
(define *showevaluationof-ignore-output* "~*")
(set! *showevaluationof-ignore-output* *showevaluationof-ignore-output*)
(define *showevaluationof-error* "
%CAUTION: error lines ~A-~A on ~A
    yields ~A
%END OF ERROR
")
(set! *showevaluationof-error* *showevaluationof-error*)

;;; This variable is used by ppfeeley.scm. It is emitted between
;;; lines. It is there so it may be redefined. Thus every
;;; pretty-printed line may be prefixed by a user-defined string.

(set! *inter-line-string* (make-string 1 #\newline))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Eval
;;; Read and evaluates a Sexpression. If an error is detected then
;;; LiSP2TeX-eval recturns a specific list. LiSP2TeX-eval is a non
;;; portable function defined in port-XXX.scm. Since there is an
;;; evaluation performed by the underlying Scheme, the expression is
;;; read with the preferred case sensibility of the underlying Scheme.

(define *eval-error-format*
  "~%LiSP2TeX (Eval) error (line ~A-~A): ~A~%original message: ~A " )
(set! *eval-error-format* *eval-error-format*)

(define (run-Eval stream)
  ;; save the current settings of the reader
  (let* ((ri *respect-input*)
         (ui *uppercase-input*)
         (li *lowercase-input*)
         (start-line (numbering-stream-line stream)) )
    ;; sets the reader to the way eval prefers
    (set! *respect-input*   *evaluator-preferred-respect-input*)
    (set! *uppercase-input* *evaluator-preferred-uppercase-input*)
    (set! *lowercase-input* *evaluator-preferred-lowercase-input*)
    (let* ((exp (numbering-read stream))
           (end-line (numbering-stream-line stream)) )
      ;; restaure the reader
      (set! *respect-input*   ri)
      (set! *uppercase-input* ui)
      (set! *lowercase-input* li)
      ;; and only now evaluates the read expression
      (let ((result (LiSP2TeX-eval exp)))
        (when (and (pair? result)
                   (eq? (car result) '***EVALUATION-ERROR***) )
          (command-error *eval-error-format*
                         start-line
                         end-line
                         exp
                         (cdr result) ) ) ) ) ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; FromFile
;;; There is a concept of current file. Whenever you use the FromFile
;;; command a null string in lieu of the filename refers to the current file.

(define *the-current-scanned-file* #f)

;;; The list of all scanned files. This is just a list of
;;; scanned-files ie a kind of records with these fields:
;;;      filename              (the name of the file)
;;;      scanned-expressions   (a vector of scanned-expressions)
;;;      string                (the entire content of the file)

(define *already-scanned-files* '())

;;; Find a scanned file and scan it (with the given scanner) if not
;;; yet done. Set it as the current file. It is left to the
;;; implementation to detect if the file does not exist since this is
;;; not possible in standard Scheme.

(define (find-scanned-file filename scanner)
  (define (lookup sfs)
    (cond ((null? sfs)
           (let ((file (search-file filename *path*)))
             (cond 
              (file 
               ;; scan file and record it:
               (command-echo "[Scan ~U" file)
               (set! *already-scanned-files*
                     (cons (scanner file)
                           *already-scanned-files* ) )
               (command-echo "]")
               (set! *the-current-scanned-file*
                     (car *already-scanned-files*) ) )
              ;; Missing file:
              (else (set! *the-current-scanned-file* #f) ) ) )
           *the-current-scanned-file* )
          ((string=? filename (scanned-file-filename (car sfs)))
           ;; check whether the scanner is correct:
           (if (eq? scanner (scanned-file-scanner (car sfs)))
               (car sfs)
               (command-error *lookupfile-error0* filename) ) )
          (else (lookup (cdr sfs))) ) )
  (lookup *already-scanned-files*) )

;;; The type of the file does not match its use. For instance, you say
;;; \FromFile("f.scm" ...) and then \FromCFile("f.scm" ...).
(define *lookupfile-error0*
  "~%LiSP2TeX (From*File) error: Syntax mismatch on file ~A" )
(set! *lookupfile-error0* *lookupfile-error0*)

;;; Find a scanned file and set up default global current variables to
;;; remember the current-scanned-file. If not found, emit a warning
;;; and return #f instead.

(define (determine-scanned-file command filename scanner)
  (if (string=? filename "")
      (if *the-current-scanned-file*
          (if (eq? scanner (scanned-file-scanner *the-current-scanned-file*))
              *the-current-scanned-file*
              (begin (command-error 
                      *lookupfile-error0* 
                      (scanned-file-filename *the-current-scanned-file*) )
                     #f ) )
          (begin (command-error *fromfile-error3* command)
                 #f ) )
      (let ((sf (find-scanned-file filename scanner)))
        (if sf 
            (set! *the-current-scanned-file* sf)
            (command-error *fromfile-error4* command filename) )
        sf ) ) )

(define *fromfile-error1* 
  "~%LiSP2TeX (~U) error: Not a filename ~A " )
(set! *fromfile-error1* *fromfile-error1*)
(define *fromfile-error2*
  "~%LiSP2TeX (~U) error: Missing arguments ~A " )
(set! *fromfile-error2* *fromfile-error2*)
(define *fromfile-error3*
  "~%LiSP2TeX (~U) error: No current scanned file ")
(set! *fromfile-error3* *fromfile-error3*)
(define *fromfile-error4*
  "~%LiSP2TeX (~U) error: Cannot find file ~A~%" )
(set! *fromfile-error4* *fromfile-error4*)

;;;;;;;;;;;;;;;;;;;;;;;;;;; Utilities

;;; Determine the line number (that is 1 + the number of preceding newlines).

(define (line-number sf limit)
  (let ((s (scanned-file-string sf)))
    (let count ((i 0)(c 1))
      (if (< i limit)
          (count (+ i 1) 
                 (if (char=? #\newline (string-ref s i))
                     (+ c 1)
                     c ) )
          c ) ) ) )

;;; compute an index (aa .. az ba .. ).

(define get-current-index-as-string
  (let ((counter (+ 25 (* 26 26))))
    (lambda ()
      (define (convert i)
        ;; assume 0 <= i < 26
        (substring "abcdefghijklmnopqrstuvwxyz" i (+ i 1)) )
      (define (gcias i)
        (let ((q (quotient i 26)))
          (if (> q 0)
              (string-append (gcias (- q 1)) (convert (modulo i 26)))
              (convert i) ) ) )
      (set! counter (+ counter 1))
      (gcias counter) ) ) )
 
;;; Check if filestring[beg:end[ is equal to string.

(define (isa-substring? string filestring beg end)
  (let ((len (string-length string)))
    (define (tryit beg)
      (or (compare 0 beg)
          (and (< beg end)
               (tryit (+ beg 1)) ) ) )
    (define (compare i j)
      ;;(display `(compare ,i ,j)) ;; DEBUG
      (if (< i len) 
          (and (< j end)
               (eqv? (string-ref string i) (string-ref filestring j))
               (compare (+ i 1) (+ j 1)) )
          #t ) )
    ;;(display `(isa-substring? ,string ,beg ,end))(newline) ; DEBUG
    (and (not (= beg end))
         (tryit beg) ) ) )

;;; The comments of the next four definitions serve as tests for 
;;; the extraction of comments. [This first block of comments should
;;; be ignored if *include-last-preceding-comment* is true.]
   
;;; This variable controls whether the last comment that precedes an
;;; expression should be extracted as well. [This comment is not 
;;; separated from the following definition by an empty-line.]
(define *include-last-preceding-comment* #t)
;;; Tell that this variable is mutable. [This comment is followed by 
;;; two empty lines that should remain here for testing purposes.]
                 
        
(set! *include-last-preceding-comment*  *include-last-preceding-comment*)

;;; [First block of comments. Just one line.]

;;; The identity and the ignorer:

(define (identity x) x)

(define (ignore string) "")

;;; This function filters a comment to keep the last block of comments.

(define (keep-last-lisp-comment comment)
  (let ((size (string-length comment))
        (last "") )
    ;; returns the index of the first character of the next line
    ;; (or the size of the string if we were already on the last line).
    (define (goto-next-line i)
      (if (< i size)
          (if (char=? (string-ref comment i) #\newline)
              (+ i 1)
              (goto-next-line (+ i 1)) )
          size ) )
    ;; Is the current line an empty line ?
    (define (empty-line? i)
      (if (< i size)
          (let ((current-char (string-ref comment i)))
            (or (char=? current-char #\newline)
                (and (memq current-char '(#\space #\tab))
                     (empty-line? (+ i 1)) ) ) )
          #f ) )
    ;; Is the current line a comment line ?
    (define (comment-line? i)
      (if (< i size)
          (let ((current-char (string-ref comment i)))
            (or (char=? current-char #\;)
                (and (memq current-char '(#\space #\tab))
                     (comment-line? (+ i 1)) ) ) )
          #f ) )
    ;; skip blank lines. Assume we are at the beginning of a line.
    (define (skip-empty-lines i)
      (if (empty-line? i)               ; INEFFICIENT
          (let ((i (goto-next-line i))) ; INEFFICIENT
            (if (< i size)
                (skip-empty-lines i)    ; INEFFICIENT
                last ) )
          (if (comment-line? i)
              (accumulate-comment-lines i i)
              last ) ) )
    ;; accumulate comment lines. Assume we are at the beginning of a line.
    (define (accumulate-comment-lines i start)
      (if (empty-line? i)
          (begin (set! last (substring comment start i))
                 (skip-empty-lines i) )
          (if (comment-line? i)
              (let ((i (goto-next-line i)))
                (if (< i size)
                    (accumulate-comment-lines i start) 
                    (substring comment start i) ) )
              (substring comment start i) ) ) )
    (skip-empty-lines 0) ) )

;;; Filter the definition.

(define *filter-lisp-definition* identity)
(set! *filter-lisp-definition*   *filter-lisp-definition*)

;;; filter the global comments.

(define (default-filter-lisp-comment comment)
  (if *include-last-preceding-comment*
      (keep-last-lisp-comment comment)
      comment ) )
(define *filter-lisp-comment* default-filter-lisp-comment)
(set! *filter-lisp-comment*   *filter-lisp-comment*)

;;; The temporary file to use

(define *tmp-file* "/tmp/lisp2tex.")
(set! *tmp-file* *tmp-file*)

(define get-tmp-file
  (let ((counter 100))
    (lambda ()
      (set! counter (+ counter 1))
      (string-append *tmp-file* (number->string counter)) ) ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Associate filename extension and snippet environments.

(define *snippet-environment* (list))

(define (define-snippet-environment envname . extensions)
  (set! *snippet-environment*
        (append (map (lambda (ext) 
                       (cons ext envname) )
                     extensions )
                *snippet-environment* ) ) )

(define (snippet-environment filename)
  (let* ((ext (extension filename))
         (value (assoc ext *snippet-environment*)) )
    (if value (cdr value) "") ) )

(define-snippet-environment "scheme" "scm" "l2t")
(define-snippet-environment "c"      "c" "h")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; FromFile
;;; Include a definition, it is reproduced as it is.

(define (run-FromFile stream)
  (let ((e (numbering-read stream)))
    (let ((options (get-options e))
          (e       (get-arguments e)) )
      (if (pair? e)
          (if (string? (car e))
              (command-FromFile options (car e) (cdr e))
              (command-error *fromfile-error1* "FromFile" e) )
          (command-error *fromfile-error2* "FromFile" e) ) ) ) )

;;; This keeps backward compatibility. 

(define (command-FromFile options filename keys)
  (if (pair? options)
      (cond ((get-option-value options "line-number")
             (new-command-FromFile options filename keys) )
            ((get-option-value options "nth")
             (nth-command-FromFile options filename keys) )
            (else (old-command-FromFile filename keys)) )
      (old-command-FromFile filename keys) ) )

;;; The old version of command-FromFile:

(define (old-command-FromFile filename keys)
  (let ((sf (determine-scanned-file "FromFile" filename scan-file)))
    (when sf
      (let ((ses (mapcan (lambda (key) (find-scanned-expression key sf))
                         keys ))
            (between? #f) )
        (for-each (lambda (se)
                    (unless (eq? se '***MISSING-EXPRESSION***)
                      (pp-format 
                       stdout-port 
                       *lisp-citation*
                       (scanned-expression-type se)
                       (scanned-expression-key se) ) ) )
                  ses )
        (when (pair? ses) (pp-format stdout-port *beginlisp-block*))
        (for-each
         (lambda (se)
           (if between?
               (pp-format stdout-port *interlisp-block*)
               (set! between? #t) )
           (if (eq? se '***MISSING-EXPRESSION***)
               (pp-format stdout-port *lisp-error-appearance*
                          `(command-FromFile ,filename ,keys) )
               (pp-format stdout-port *lisp-appearance*
                          (*filter-lisp-comment*
                           (substring (scanned-file-string sf)
                                      (scanned-expression-start se)
                                      (scanned-expression-real-start se) ) )
                          (*filter-lisp-definition*
                           (substring (scanned-file-string sf)
                                      (scanned-expression-real-start se)
                                      (scanned-expression-end se) ) ) ) ) )
         ses )
        (when (pair? ses) (pp-format stdout-port *endlisp-block*)) ) ) ) )

(define *lisp-citation*   "\\snippetCite{~U}{~U}~%")
(set! *lisp-citation* *lisp-citation*)
(define *beginlisp-block* "\\snippet[scheme]~%")
(set! *beginlisp-block* *beginlisp-block*)
(define *interlisp-block* " ~%")
(set! *interlisp-block* *interlisp-block*)
(define *endlisp-block*   " \\endsnippet ")
(set! *endlisp-block* *endlisp-block*)
(define *lisp-appearance* "~U~U")
(set! *lisp-appearance* *lisp-appearance*)
(define *lisp-error-appearance* 
  "* ** *** **** MISSING EXPRESSION ~S **** *** ** *")
(set! *lisp-error-appearance* *lisp-error-appearance*)

;;; The old version of command-FromFile adapted so it extracts only
;;; a given definition from a file. Expressions are numbered from 0.

(define (nth-command-FromFile options filename keys)
  (let ((n* (let ((options (get-option-value options "nth")))
              (cond 
               ((pair? options)   options)
               ((number? options) (list options))
               (else (command-error *nth-command-fromfile-error1* options)) ) ))
        (sf (determine-scanned-file "FromFile" filename scan-file)) )
    (when sf
      (let* ((ses (mapcan (lambda (key) (find-scanned-expression key sf))
                          keys ))
             (ses (let loop ((n* n*))
                    (if (pair? n*)
                        (if (and (number? (car n*))
                                 (<= 0 (car n*))
                                 (< (car n*) (length ses)) )
                            (cons (list-ref ses (car n*))
                                  (loop (cdr n*)) )
                            (begin 
                              (command-error *nth-command-fromfile-error2* 
                                             (car n*) )
                              (loop (cdr n*)) ) )
                        '() ) ))
             (between? #f) )
        (for-each (lambda (se)
                    (unless (eq? se '***MISSING-EXPRESSION***)
                      (pp-format 
                       stdout-port 
                       *lisp-citation*
                       (scanned-expression-type se)
                       (scanned-expression-key se) ) ) )
                  ses )
        (when (pair? ses) (pp-format stdout-port *beginlisp-block*))
        (for-each
         (lambda (se)
           (if between?
               (pp-format stdout-port *interlisp-block*)
               (set! between? #t) )
           (if (eq? se '***MISSING-EXPRESSION***)
               (pp-format stdout-port *lisp-error-appearance*
                          `(command-FromFile ,filename ,keys) )
               (pp-format stdout-port *lisp-appearance*
                          (*filter-lisp-comment*
                           (substring (scanned-file-string sf)
                                      (scanned-expression-start se)
                                      (scanned-expression-real-start se) ) )
                          (*filter-lisp-definition*
                           (substring (scanned-file-string sf)
                                      (scanned-expression-real-start se)
                                      (scanned-expression-end se) ) ) ) ) )
         ses )
        (when (pair? ses) (pp-format stdout-port *endlisp-block*)) ) ) ) )

(define *nth-command-fromfile-error1* 
  "~%LiSP2TeX: Incorrect option value (not a (or list of) number) ~A~%" )
(set! *nth-command-fromfile-error1* *nth-command-fromfile-error1*)
(define *nth-command-fromfile-error2* 
  "~%LiSP2TeX: No definition numbered ~A! ~%" )
(set! *nth-command-fromfile-error2* *nth-command-fromfile-error2*)

;;; New version 
                  
(define (new-command-FromFile options filename keys)
  (let ((sf (determine-scanned-file "FromFile" filename scan-file)))
    (if sf
        (let ((ses (mapcan (lambda (key) (find-scanned-expression key sf))
                           keys )) )
          (generic-outputter "FromFile" options filename keys sf ses 
                             "scheme"
                             *citation*
                             *begin-block*
                             *inter-block*
                             *error-appearance*
                             *appearance*
                             *filter-lisp-comment*
                             *filter-lisp-definition*
                             *end-block* ) ) ) ) )

;;; Prefix all snippets with a series of citations and lines numbers.
;;; Presently if works for FromFile and ExtractFromFile that have
;;; comparable scanned-expressions but not for FromCFile (whose
;;; scanned-c-expression lack real-start field). 

(define (generic-outputter command options filename keys sf ses 
                           snippetenv
                           citation-format
                           begin-format
                           inter-format
                           error-format
                           snippet-format
                           filter-comment
                           filter-definition
                           end-format )
  (let ((between? #f)
        (indexes (map (lambda (se) (get-current-index-as-string))
                      ses )) )
    (when (get-option-value options "line-number")
      (pp-format stdout-port "\\begingroup\\snippetWithLinestrue~%") )
    ;; generate the \snippetCite and \setSnippetLineNumber:
    (for-each
     (lambda (se index)
       (unless (eq? se '***MISSING-EXPRESSION***)
         (pp-format
          stdout-port 
          citation-format
          (scanned-expression-type se)
          (scanned-expression-key se)
          index
          (line-number sf (scanned-expression-start se))
          (line-number sf (scanned-expression-real-start se))
          (line-number sf (scanned-expression-end se)) ) ) )
     ses indexes )
    ;; Start to insert the extracted snippets with a prologue, an
    ;; interlude between any two snippets, and an epilogue.
    (when (and (pair? ses) (pair? indexes))
      (pp-format stdout-port begin-format snippetenv (car indexes)) )
    (for-each
     (lambda (se index)
       (if between?
           (pp-format stdout-port inter-format index)
           (set! between? #t) )
       (if (eq? se '***MISSING-EXPRESSION***)
           (pp-format stdout-port error-format
                      `(,command ,options ,filename ,keys) )
           (pp-format stdout-port snippet-format
                      (filter-comment
                       (substring (scanned-file-string sf)
                                  (scanned-expression-start se)
                                  (scanned-expression-real-start se) ) )
                      (filter-definition
                       (substring (scanned-file-string sf)
                                  (scanned-expression-real-start se)
                                  (scanned-expression-end se) ) ) ) ) )
     ses indexes )
    (when (pair? ses) (pp-format stdout-port end-format))
    (when (get-option-value options "line-number")
      (pp-format stdout-port "\\endgroup~%") ) ) )

(define *citation*
  "\\snippetCite{~A}{~A}~%\\setSnippetLineNumber{~U}{~A,~A,~A}~%")
(set! *citation* *citation*)
(define *begin-block* "\\snippet[~U]\\snippetLineNumber~U~%")
(set! *begin-block* *begin-block*)
(define *inter-block* " \\snippetLineNumber~U~%")
(set! *inter-block* *inter-block*)
(define *end-block*   " \\endsnippet ")
(set! *end-block* *end-block*)
(define *appearance* "~U~U")
(set! *appearance* *appearance*)
(define *error-appearance* 
  "* ** *** **** MISSING EXPRESSION ~S **** *** ** *")
(set! *error-appearance* *error-appearance*)

;;; Find the expressions matched by key in scanned-file.
;;; A key can be  
;;;    a symbol : find THE Sexpression that defines this name,
;;;    a pair (considered as a pattern): find ALL Sexpressions matched.

(define (find-scanned-expression key scanned-file)
  (let* ((ses (scanned-file-scanned-expressions scanned-file))
         (found #f)
         (max (vector-length ses)) )
    (define (lookup i)
      (if (< i max)
          (let ((se (vector-ref ses i)))
            (cond ((and (symbol? key)
                        (scanned-expression-key se)
                        (string-ci=? (symbol->string key)
                                     (scanned-expression-key se) ) )
                   (set! found #t)
                   (list se) )
                  ((and (pair? key)
                        (naive-match (scanned-expression-original se) key) )
                   (set! found #t)
                   (cons se (lookup (+ 1 i))) )
                  (else (lookup (+ 1 i))) ) )
          (if found
              '()
              (begin
                (command-error *find-definition-error* key)
                '(***MISSING-EXPRESSION***) ) ) ) )
    (lookup 0) ) )

(define *find-definition-error*
  "~%LiSP2TeX (*From*File) error: Cannot find definition matching ~A " )
(set! *find-definition-error* *find-definition-error*)

;;; This naive pattern matcher recognizes ?- and ... at any level.
;;; ?- matches any Sexp while ... matches any sequence of Sexp. This
;;; latter can only be used in the context of a list.
;;; CommonLisp does not accept ... as a valid symbol so use this periphrase;
;;; Users of the pseudo port would probably redefine ... to be an other thing
;;; for instance ??- or ???.
;;; (-or- pattern pattern .. pattern) matches any expression matching one of
;;; these patterns.

(define *mono-Sexp-pattern* '?-)
(set! *mono-Sexp-pattern* *mono-Sexp-pattern*)
(define *poly-Sexp-pattern* (string->symbol "..."))
(set! *poly-Sexp-pattern* *poly-Sexp-pattern*)
(define *or-Sexp-pattern* '-or-)
(set! *or-Sexp-pattern* *or-Sexp-pattern*)

(define (naive-match expression pattern)
  (define (naive-match-list expressions patterns)
    (if (pair? patterns)
        (if (eq-ci? (car patterns) *poly-Sexp-pattern*)
             ; accepts any sequence of things
            (or (naive-match-list expressions (cdr patterns))
                (and (pair? expressions)
                     (naive-match-list (cdr expressions) patterns) ) )
            (and (pair? expressions)
                 (naive-match (car expressions) (car patterns))
                 (naive-match-list (cdr expressions) (cdr patterns)) ) )
        (naive-match expressions patterns) ) )
  (or (eq-ci? pattern *mono-Sexp-pattern*)              ; accepts anything
      (if (eq-ci? pattern *poly-Sexp-pattern*)
          (command-error *match-dotdotdot-error*
                         *poly-Sexp-pattern* pattern )
          (if (pair? pattern)
              (if (eq-ci? (car pattern) *or-Sexp-pattern*)
                  (let enum ((pats (cdr pattern)))
                    (if (pair? pats)
                        (or (naive-match expression (car pats))
                            (enum (cdr pats)) )
                        #f ) )
                  (naive-match-list expression pattern) )
              (equal? expression pattern) ) ) ) )

(define *match-dotdotdot-error*
  "Bad use of ~A in pattern" )
(set! *match-dotdotdot-error* *match-dotdotdot-error*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; FromCFile
;;; Include a definition from a C file, it is reproduced as it is except
;;; that backslashes are transformed in the extracted fragment.

(define (run-FromCFile stream)
  (let ((e (numbering-read stream)))
    (let ((options (get-options e))
          (e       (get-arguments e)) )
      (if (pair? e)
          (if (string? (car e))
              (command-FromCFile options (car e) (cdr e))
              (command-error *fromfile-error1* "FromCFile" e) )
          (command-error *fromfile-error2* "FromCFile" e) ) ) ) ) 

(define (command-FromCFile options filename keys)
  (let ((sf (determine-scanned-file "FromFile" filename scan-c-file)))
    (if sf
        (let ((ses (mapcan (lambda (key) (find-scanned-c-expression key sf))
                           keys ))
              (between? #f) )
          (when (pair? ses) (pp-format stdout-port *beginc-block*))
          (for-each 
           (lambda (se)
             (if between?
                 (pp-format stdout-port *interc-block*)
                 (set! between? #t) )
             (if (eq? se '***MISSING-EXPRESSION***)
                 (pp-format stdout-port *c-error*
                            (symbol->string se) )
                 (pp-format stdout-port *c-block*
                            (c-substring (scanned-c-file-string sf) se) ) ) )
                    ses )
          (when (pair? ses) (pp-format stdout-port *endc-block*)) ) ) ) )

(define *beginc-block* "\\snippet[c]~%")
(set! *beginc-block* *beginc-block*)
(define *interc-block* " ~%")
(set! *interc-block* *interc-block*)
(define *c-error* "~U")
(set! *c-error* *c-error*)
(define *c-block* "~U")
(set! *c-block* *c-block*)
(define *endc-block*   " \\endsnippet ")
(set! *endc-block* *endc-block*)

;;; Extract a substring but, It is possible to ask to double backslah
;;; for TeX, and leave C comments as they are.

(define *double-backslash* #f)
(set! *double-backslash* *double-backslash*)
(define *omit-comments* #f)
(set! *omit-comments* *omit-comments*)

(define (c-substring string se)
  (let* ((beg (scanned-c-expression-start se))
         (end (scanned-c-expression-end se))
         (comments (if *omit-comments* 
                       (scanned-c-expression-comments se)
                       '() ))
         (result (make-extensible-string (- end beg))) )
    (do ((i beg (+ i 1)))
        ((>= i end))
      (cond
       ;; skip comments if any.
       ((and (pair? comments)
             (= i (car (car comments))) )
        (set! i (- (cdr (car comments)) 1))
        (set! comments (cdr comments)) )
       (else
        (let ((ch (string-ref string i)))
          (if (and *double-backslash* (eqv? ch #\\))
              (increment-string ch result) )
          (increment-string ch result) ) ) ) )
    (finalize-string result) ) )

;;; Find the expressions matched by key in scanned-file.  A key can be
;;; a symbol or a string: return all the expressions containing that
;;; string.

(define (find-scanned-c-expression key scanned-c-file)
  (let* ((sf  scanned-c-file)
         (ses (scanned-file-scanned-expressions sf))
         (key (cond ((symbol? key) (symbol->string key))
                    ((string? key) key)
                    (else (command-error *fromcfile-error5* key)) ))
         (max (vector-length ses)) )
    (define (lookup i result)
      (if (< i max)
          (let ((se (vector-ref ses i)))
            (cond ((if (and (= (scanned-c-expression-bodystart se) 0)
                            (= (scanned-c-expression-bodyend se) 0) )
                       (isa-substring? key
                                       (scanned-c-file-string sf)
                                       (scanned-c-expression-start se)
                                       (scanned-c-expression-end se) )
                       (or (isa-substring? key
                                           (scanned-c-file-string sf)
                                           (scanned-c-expression-start se)
                                           (scanned-c-expression-bodystart se) )
                           (isa-substring? key
                                           (scanned-c-file-string sf) 
                                           (scanned-c-expression-bodyend se)
                                           (scanned-c-expression-end se) ) ) )
                   (lookup (+ 1 i) (cons se result))
                   )
                  (else (lookup (+ 1 i) result)) ) )
          result ) )
    (let ((r (lookup 0 '())))
      (if (pair? r) (reverse r)
          (begin
            (command-error *find-definition-error* key)
            '(***MISSING-EXPRESSION***) ) ) ) ) )

(define *fromcfile-error5* 
  "~%LiSP2TeX (FromCFile) error: Incorrect key ~A~%" ) 
(set! *fromcfile-error5* *fromcfile-error5*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ExtractFromFile 

(define (run-ExtractFromFile stream)
  (let ((e (numbering-read stream)))
    (let ((options (get-options e))
          (e       (get-arguments e)) )
      (if (pair? e)
          (if (string? (car e))
              (command-ExtractFromFile options (car e) (cdr e))
              (command-error *fromfile-error1* "ExtractFromFile" e) )
          (command-error *fromfile-error2* "ExtractFromFile" e) ) ) ) )

(define (command-ExtractFromFile options filename keys)
  (let ((sf (determine-scanned-file "ExtractFromFile" 
                                    filename 
                                    scan-external-file )))
    (if sf
        (let ((ses (mapcan (lambda (key)
                             (find-scanned-external-expression key sf) )
                           keys )) )
          (generic-outputter "ExtractFromFile" options filename keys sf ses 
                             (snippet-environment filename)
                             *citation*
                             *begin-block*
                             *inter-block*
                             *error-appearance* 
                             *appearance*
                             *filter-external-comment* 
                             *filter-external-body*
                             *end-block* ) ) ) ) )

(define *filter-external-comment* ignore)
(set! *filter-external-comment* *filter-external-comment*)
(define *filter-external-body* identity)
(set! *filter-external-body* *filter-external-body*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Utility functions

;;; These options allow to display the progress of the LiSP2TeX filter.
;;; Progress is reported on stderr not to be confused with the regular
;;; output.

(define *verbose* #f)
(set! *verbose* *verbose*)
(define *ultra-verbose* #f)
(set! *ultra-verbose* *ultra-verbose*)

;;; Reports some progress
(define (command-echo fmt . args)
  (when *verbose*
    (LiSP2TeX-apply pp-format stderr-port fmt args)
    (flush-all-buffers) ) )

;;; reports some error on both streams (to see it in the output TeX)
(define (command-error fmt . arguments)
  (set! *exit-status* 1)
  (display-error fmt arguments)
  (flush-all-buffers)
  (if *stop-at-first-error*
      (begin 
        (display-error *abort-format* arguments)
        (*exit*) ) ) )

(define *abort-format*
  "LiSP2TeX stopped at first error: ~*~%" )
(set! *abort-format* *abort-format*)

(define (*exit*)
  (LiSP2TeX-exit *exit-status*) )

;;; Default behavior for display-error.
(define (display-error fmt arguments)
  (LiSP2TeX-apply pp-format stdout-port fmt arguments)
  (LiSP2TeX-apply pp-format stderr-port fmt arguments) )

;;; display-error should  be defined somewhere in port-*.scm
;(define display-error display-error-default)

;;; The list of directories where to search (customization or regular) files. 
;;; Only the current directory by default.
(define *path* '("."))
(set! *path* *path*)

;;; Emit a header (a TeX comment) identifying the version of LiSP2TeX.
;;; It is output on stdout as a TeX comment.

(define *header-format*
  "%%% LiSP2TeX (~U) ~U, Copyright (C) 2000 <Christian.Queinnec@lip6.fr>
%%% LiSP2TeX comes with ABSOLUTELY NO WARRANTY; for details see COPYING file.
%%% This is free software, and you are welcome to redistribute it
%%% under certain conditions; see COPYING file again.
%%%
%%% This file was automatically produced by LiSP2TeX (~U) ~U
%%% with arguments: ~S
%%%                   --- PLEASE DO NOT EDIT --- ~%" )
(set! *header-format* *header-format*)

(define (emit-header args)
  (pp-format stdout-port *header-format* 
             scm-name version 
             scm-name version
             args ) )

;;; Identify the current version of LiSP2TeX.
(define (emit-version-header outport)
  (pp-format 
   outport
   "LiSP2TeX (~U) ~U [Christian.Queinnec@lip6.fr]~%" 
   scm-name version ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; exit
;;; Sometimes it is useful to stop processing. This may be done by
;;; calling the LiSP2TeX-exit function. That function is defined
;;; in the port-*.scm files.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; MAIN
;;; The main entry point, handles options then filters stdin.
;;; Scheme->C calls it from the shell with the shell arguments in args.

(define *no-option*
  "~%LiSP2TeX option error: empty option " )
(set! *no-option* *no-option*)
(define *too-much---options*
  "~%LiSP2TeX option error: too much -- options " )
(set! *too-much---options* *too-much---options*)
(define *empty-option*
  "~%LiSP2TeX option error: No option after dash" )
(set! *empty-option* *empty-option*)
(define *cannot-find-file*
  "~%LiSP2TeX option error: Cannot find file ~A " )
(set! *cannot-find-file* *cannot-find-file*)
(define *unknown-option*
  "~%LiSP2TeX option error: Unrecognized option: ~A " )
(set! *unknown-option* *unknown-option*)

(define *exit-status* 0)

(define (LiSP2TeX args)
  (call/cc
   (lambda (return)
     (set! *exit* (lambda () (return 'done)))
     ;; Analyse options
     (let ((does--appear? #f)
           (settings '())
           (files '())
           (original-args args) )
       (do ((args args (cdr args)))
           ((null? args) #f)
         (cond
          ;; refuses empty options
          ((= 0 (string-length (car args)))
           (command-error *no-option*) )
          ;; outputs the version identifier
          ((equal? (car args) "--version")(emit-version-header stderr-port))
          ;; separates the parameter files from the files to filter
          ((equal? (car args) "--")
           (cond (does--appear? 
                  (command-error *too-much---options*) )
                 (else
                  (set! does--appear? #t)
                  (set! settings files)
                  (set! files '()) ) ) )
          ;; A real option (or combination of options)
          ((char=? (string-ref (car args) 0) #\-)
           (if (> (string-length (car args)) 1)
               (analyze-option (car args) 1)
               (command-error *empty-option*) ) )
          (else
           (set! files (cons (car args) files)) ) ) )
       ;; Reverse the order of directories pushed onto *path*
       (set! *path* (reverse *path*))
       ;; Identifies the current version of LiSP2TeX on stdout
       (emit-header original-args)
       ;; Load the parameter files. They use *path* 
       (when does--appear?
         (for-each (lambda (filename) 
                     (let ((file (search-file filename *path*)))
                       (if file 
                           (begin
                             (command-echo "[Read customization ~U..."
                                           file )
                             (load-parameter-file file)
                             (command-echo "]~%") )
                           (command-error *cannot-find-file*
                                          filename ) ) ) )
                   (reverse! settings) ) )
       ;; Now handle files to filter or the standard input
       (if (pair? files)
           ;; Handles all mentionned files
           (for-each (lambda (filename)
                       (let ((file (search-file filename *path*))
                             (old-path *path*) )
                         (if file
                             (call-with-input-file
                                 file
                               (lambda (in)
                                 ;; required by Titou Durand: add the
                                 ;; directory of the currently processed
                                 ;; file to the *path*.
                                 (set! *path* (cons (dirname file) *path*))
                                 (command-echo "[Process ~U..." filename)
                                 (look (create-numbering-stream in)
                                       numbering-stream-read-char )
                                 (command-echo "]~%")
                                 (set! *path* old-path) ) )
                             (command-error *cannot-find-file*
                                            filename ) ) ) )
                     (reverse! files) )
           ;; Handles stdin
           (look (create-numbering-stream stdin-port)
                 numbering-stream-read-char ) ) ) ) )
  ;; Terminates.
  (LiSP2TeX-exit *exit-status*) )

;;; Tell whether to stop at first evaluation error:

(define *stop-at-first-error* #f)

;;; analyze options (options start with "-")

(define (analyze-option arg i)
  (when (< i (string-length arg))
        (cond 
         ;; Add a directory to search
	 ((char=? (string-ref arg i) #\I)
	  (set! *path* (cons (substring arg 
					(+ 1 i)
					(string-length arg) )
			     *path* )) )
         ;; be verbose
	 ((char=? (string-ref arg i) #\v)
	  (set! *verbose* #t)
	  (analyze-option arg (+ 1 i)) )
         ;; be more verbose if possible
	 ((char=? (string-ref arg i) #\V)
	  (set! *ultra-verbose* #t)
	  (set! *verbose* #t)
	  (analyze-option arg (+ 1 i)) )
         ;; converts what is read to uppercase
	 ((char=? (string-ref arg i) #\U)
          (set! *respect-input*    #f)
	  (set! *lowercase-input*  #f)
	  (set! *uppercase-input*  #t) )
         ;; converts what is printed to uppercase
	 ((char=? (string-ref arg i) #\u)
          (set! *respect-output*   #f)
	  (set! *lowercase-output* #f)
	  (set! *uppercase-output* #t) )
	 ((char=? (string-ref arg i) #\L)
          (set! *respect-input*    #f)
	  (set! *lowercase-input*  #t)
	  (set! *uppercase-input*  #f) )
	 ((char=? (string-ref arg i) #\l)
          (set! *respect-output*   #f)
	  (set! *lowercase-output* #t)
	  (set! *uppercase-output* #f) )
	 ((char=? (string-ref arg i) #\=)
          (set! *respect-input*    #t)
	  (set! *lowercase-input*  #f)
	  (set! *uppercase-input*  #f)
          (set! *respect-output*   #t)
	  (set! *lowercase-output* #f)
	  (set! *uppercase-output* #f) )
         ;; Do not emit header
	 ((char=? (string-ref arg i) #\h)
	  (set! emit-header (lambda (args) #f)) )
         ;; Stop at first error
         ((char=? (string-ref arg i) #\a)
          (set! *stop-at-first-error* #t) )
	 (else 
	  (command-error *unknown-option*
			 (string-ref arg i) ) ) ) ) )

;;; end of command.scm
