;;; $Id: emit.scm,v 1.11 2004/04/08 17:35:14 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file contains the printing routines.

;;; The output base for numbers
(define *output-base* 10)
(set! *output-base* *output-base*)
(define *number-images*
  (vector #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\A #\B #\C #\D #\E #\F) )
(set! *number-images* *number-images*)

;;; The length of the line to pretty-print
(define *pretty-line-width* 60)
(set! *pretty-line-width* *pretty-line-width*)

;;; How boolean values are printed:
(define *true-image* "#T")
(set! *true-image* *true-image*)
(define *false-image* "#F")
(set! *false-image* *false-image*)

;;; Error formats
(define *unprintable-object* 
  "Do not know how to print this object: ~A" )
(set! *unprintable-object* *unprintable-object*)
(define *too-big-format*
  "Too less arguments: \"~A\"." )
(set! *too-big-format* *too-big-format*)
(define *incorrect-format*
  "Incorrect format specifier in format \"~A\", at index ~A." )
(set! *incorrect-format* *incorrect-format*)
(define *too-short-format*
  "Too much arguments for format: \"~A\"." )
(set! *too-short-format* *too-short-format*)
(define *not-a-string* 
  "format ~~U in \"~A\", not a string ~A." )
(set! *not-a-string* *not-a-string*)

;;; Thee main printing function
(define (pp-format stream fmt . arguments)
  (define (prin-dotted-list o)
    (display-string " . " stream)
    (princ o) )
  (define (princ-list o)
    (cond ((null? o) o)
          ((pair? o) 
           (if (memq (car o) '(quote quasiquote unquote unquote-splicing))
               (prin-dotted-list o)
               (begin (write-char #\space stream)
                      (princ (car o))
                      (princ-list (cdr o)) ) ) )
          (else (prin-dotted-list o)) ) )
  (define (princ-normal-list o)
    (write-char #\( stream)
    (princ (car o))
    (princ-list (cdr o))
    (write-char #\) stream) )
  (define (princ-special-list o)
    (case (car o)
      ((quote) 
       (cond ((and (pair? (cdr o)) (null? (cddr o)))
              (write-char #\' stream)
              (princ (cadr o)) )
             (else (princ-normal-list o)) ) )
      ((quasiquote) 
       (cond ((and (pair? (cdr o)) (null? (cddr o)))
              (write-char #\` stream)
              (princ (cadr o)) )
             (else (princ-normal-list o)) ) )
      ((unquote)  
       (cond ((and (pair? (cdr o)) (null? (cddr o)))
              (write-char #\, stream)
              (princ (cadr o)) )
             (else (princ-normal-list o)) ) )
      ((unquote-splicing)  
       (cond ((and (pair? (cdr o)) (null? (cddr o)))
              (display-string ",@" stream)
              (princ (cadr o)) )
             (else (princ-normal-list o)) ) )
      (else (princ-normal-list o)) ) )
  (define (princ-vector o leng)
    (do ((i 0 (+ i 1))
         (j 1 (+ j 1)) )
        ((= i leng) o)
      (princ (vector-ref o i))
      (unless (= j leng) (write-char #\space stream)) ) )
  (define (princ o)
    (cond ((symbol? o) 
           (let ((s (symbol->string o)))
             (cond 
              (*respect-output* (display-string s stream))
              (*uppercase-output* (display-string (string-uppercase s) stream))
              (*lowercase-output* (display-string (string-lowercase s) stream))
              ) ) )
          ((null? o) (display-string "()" stream))
          ;; ports are pairs in Scheme->C, so test them before
          ((input-port? o) (display-string "#<an input Port>" stream))
          ((output-port? o) (display-string "#<an output Port>" stream))
          ((pair? o) (princ-special-list o))
          ((integer? o) (prin-integer o stream))
          ((string? o) (current-write-string o stream))
          ((vector? o) (display-string "#(" stream)
                       (princ-vector o (vector-length o))
                       (display-string ")" stream) )
          ((boolean? o) 
           (display-string (if o *true-image* *false-image*) stream) )
          ((procedure? o) (display-string "#<a Function>" stream))
          ((number? o) (display-number o stream))
          ((char? o) (display o stream))
          (else ;;(command-error *unprintable-object* o)
                ; Don't report an error!
                (display o stream)
                ) ) )
  (define format-length (string-length fmt))
  (define current-write-string display-string)
  (do ((i 0 (+ 1 i)))
      ((>= i format-length) #t)
    (if (char=? (string-ref fmt i) #\~)
        (begin (set! i (+ 1 i))
               (cond 
                ;; Displays a Sexpression
                ((char-ci=? (string-ref fmt i) #\A)
                 (set! current-write-string display-string)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments))
                       (princ o) )
                     (command-error *too-big-format* fmt) ) )
                ;; Writes a Sexpression
                ((char-ci=? (string-ref fmt i) #\S)
                 (set! current-write-string write-string)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments))
                       (princ o) )
                     (command-error *too-big-format* fmt) ) )
                ;; As ~a except that the value is quoted only when needed:
                ((char=? (string-ref fmt i) #\1)
                 (set! current-write-string write-string)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments))
                       (if (or (number? o) (char? o) (boolean? o)
                               (vector? o) (string? o) )
                           (princ o)
                           (begin 
                             (write-char #\' stream)
                             (princ o) ) ) )
                     (command-error *too-big-format* fmt) ) )
                ;; pretty-greek-prints a Sexpression
                ((char-ci=? (string-ref fmt i) #\Q)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments))
                       (pp-greekify o stream) )
                     (command-error *too-big-format* fmt) ) )
                ;; glorify a Sexpression (with Marc Feeley's prettyprinter)
                ((char-ci=? (string-ref fmt i) #\G)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments))
                       ;; (pretty-print o stream)
                       (generic-write o #f *pretty-line-width* 
                                      (lambda (s) (display s stream) #t) ) )
                     (command-error *too-big-format* fmt) ) )
                ;; Displays a tilde
                ((char=? (string-ref fmt i) #\~)
                 (write-char #\~ stream) )
                ;; Displays a string without surrounding quotes
                ;; [Kept for backward compatibility but obsolete].
                ((char-ci=? (string-ref fmt i) #\U)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments))
                       (if (string? o)
                           (display-string o stream)
                           (command-error *not-a-string* fmt o) ) )
                     (command-error *too-big-format* fmt) ) )
                ;; Swallow an expression and do not print it at all
                ;; This is needed for imposed format that must receive
                ;; some values and do not want to show them.
                ((char-ci=? (string-ref fmt i) #\*)
                 (if (pair? arguments)
                     (let ((o (car arguments)))
                       (set! arguments (cdr arguments)) )
                     (command-error *too-big-format* fmt) ) )
                ;; emits a newline
                ((char=? (string-ref fmt i) #\%)
                 (newline stream) )
                (else (command-error *incorrect-format* fmt i)) ) )
        (write-char (string-ref fmt i) stream) ) )
  (or (null? arguments)
      (command-error *too-short-format* fmt) ) )
                            
(define (write-string string stream)
  (write string stream) )
(define (display-string string stream)
  (display string stream) )

(define (prin-integer n stream)
  (when (< n 0) (write-char #\- stream)
                (set! n (- n)) )
  (let ((p (quotient n *output-base*)))
    (when (> p 0) (prin-integer p stream))
    ;; changed for Gambit that produces an inexact there:
    (let ((r (inexact->exact (remainder n *output-base*))))
      (write-char (vector-ref *number-images* r)
                  stream ) ) ) )

;;; Other numbers are printed by the native Scheme
(define (display-number o stream)
  (display o stream) )

;;; Tests
; (pp-format stdout-port "foo")
; (pp-format stdout-port "foo~A bar" 1)
; (pp-format stdout-port "foo~A bar~A~%,~A." 1 2 3)
; (pp-format stdout-port "foo~A bar" -123)
; (pp-format stdout-port "foo~A bar" #t)
; (pp-format stdout-port "foo~A bar" #f)
; (pp-format stdout-port "foo~A bar" stderr-port)
; (pp-format stdout-port "foo~A bar" "STrInG")
; (pp-format stdout-port "foo~A bar" '#(1 -2 "FoO"))
; (pp-format stdout-port "foo~A bar" 'foo)
; (pp-format stdout-port "foo~A bar" '(foo 1))
; (pp-format stdout-port "foo~A bar" '(foo 1 (3 . \Ba\R)))
; (pp-format stdout-port "foo~A bar" ''foo)
; (pp-format stdout-port "foo~A bar" '(foo `(bar ,foo ,@hux) . 3))
; (pp-format stdout-port "foo~A bar" '(foo `(bar ,foo . ,hux) . 'gazonk))

;;; Erroneous tests
; (pp-format stdout-port "foo~Z bar")
; (pp-format stdout-port "foo~A bar" )
; (pp-format stdout-port "foo~A bar" 1 2)
; (pp-format stdout-port "foo~A bar" '(quote a b))

;;; end of emit.scm
