;;; $Id: find-string.scm,v 1.6 2000/10/29 16:36:12 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file defines how directives are looked for in the standard input.

;;; PORT note: Macros are still not portable so an alternative is
;;; to use the `look.scm' file instead of this one. The `look.scm' file
;;; is the expansion of this one and does not contain any macros.

;;; This is the letter by which directives start. It is then easy to 
;;; modify it to become #\@ as for TeXinfo.

(define *directive-sign* #\\)
(set! *directive-sign* *directive-sign*)

;;; A (big) macro to define the function to look for the directives to
;;; be translated. The generated pattern is not optimal, I have in the 
;;; mind a kind of Boyer Moore with multiple strings that I will
;;; eventually write.

(define-lisp2tex-macro (generate-look-function defs)
  (define (string . chars)              ; Curiously lacks from Scheme->C
    (let* ((size (length chars))
           (string (make-string size)) )
      (do ((i 0 (+ 1 i))
           (chars chars (cdr chars)) )
          ((= i size) string)
        (string-set! string i (car chars)) ) ) )
  (define gensym 
    (let ((counter 100))
      (lambda (string)
        (set! counter (+ 1 counter))
        (string->symbol
         (string-append string (number->string counter)) ) ) ) )
  (define (reverse! l)
    (define (nreverse l r)
      (if (pair? l)
          (let ((cdrl (cdr l)))
            (set-cdr! l r)
            (nreverse cdrl l) )
          r ) )
    (nreverse l '()) )
  ;; Analysis of the commands
  (let ((read-char (gensym (symbol->string 'read-char)))
        (port      (gensym (symbol->string 'port))) )
    ;; augment a string with a character:
    (define (string-add-char string char)
      (let* ((size (string-length string))
             (newstring (make-string (+ 1 size))) )
        (do ((i 0 (+ 1 i)))
            ((= i size) 
             (string-set! newstring i char)
             newstring )
          (string-set! newstring i (string-ref string i)) ) ) )
    ;; Actually, this is string->list:
    (define (string->list-of-characters string)
      (let ((size (string-length string)))
        (do ((i 0 (+ 1 i))
             (result '()) )
            ((= i size) (reverse! result))
          (set! result (cons (string-ref string i) result)) ) ) )
    ;; returns the list of characters from chars without duplication:
    (define (no-duplicates chars)
      (if (null? chars) '()
          (if (memv (car chars) (cdr chars))
              (no-duplicates (cdr chars))
              (cons (car chars) (no-duplicates (cdr chars))) ) ) )
    (define (generate ch defs prefix)
      `(cond
        ((eof-object? ,ch) 
         (write-char *directive-sign* stdout-port)
         ,@(if (not (string=? prefix ""))
               `((display ,prefix stdout-port))
               `() ) )
        ,@(map (lambda (char) 
                 (generate-cond-clause ch char defs prefix) )
               (no-duplicates (map (lambda (def) (car (car def)))
                                   defs )) )
        (else (write-char *directive-sign* stdout-port)
              ,@(if (not (string=? prefix ""))
                    `((display ,prefix stdout-port))
                    `() )
              (look ,ch) ) ) )
    ;; partially evaluated 'generate' for the first directive character:
    (define (generate1 ch defs)
      `(cond
        ((eof-object? ,ch) 'nothing)
        ,(generate-cond-clause1 ch defs)
        (else (write-char ,ch stdout-port)
              (look (,read-char ,port)) ) ) )
    ;; partially evaluated 'generate-cond-clause' 
    (define (generate-cond-clause1 ch defs)
      `((char=? ,ch *directive-sign*)
        ,(let ((ch (gensym (symbol->string 'char))))
           (cond                  
            ((and (= (length defs) 1)
                  (null? (car (car defs))) )
             `(begin
                (when *ultra-verbose* 
                  (command-echo ,(string-add-char prefix char)) )
                (,(cadr (car defs)) ,port)
                (flush-all-buffers)
                (look (,read-char ,port)) ) )
            (else
             `(let ((,ch (,read-char ,port)))
                ,(generate ch defs "") ) ) ) ) ) )
    (define (generate-cond-clause ch char defs prefix)
      `((char=? ,ch ,char)
        ,(let ((defs (do ((defs defs (cdr defs))
                          (result '()) )
                         ((null? defs) result)
                       (if (char=? char (caar (car defs)))
                           (set! result (cons (cons (cdar (car defs))
                                                    (cdr (car defs)) )
                                              result )) ) ))
               (ch (gensym (symbol->string 'char))) )
           (cond                  
            ((and (= (length defs) 1)
                  (null? (car (car defs))) )
             `(begin 
                (when *ultra-verbose* 
                  (command-echo ,(string-add-char prefix char)) )
                (,(cadr (car defs)) ,port)
                (flush-all-buffers)
                (look (,read-char ,port)) ) )
            (else
             `(let ((,ch (,read-char ,port)))
                ,(generate
                  ch
                  defs
                  (string-add-char prefix char) ) ) ) ) ) ) )
    ;; The body:
    (let* ((ch (gensym (symbol->string 'char)))
           (defs (map (lambda (def)
                        (cons (string->list-of-characters (car def))
                              (cdr def) ) )
                      defs ))
           (result `(define (look ,port ,read-char)
                      (define (look ,ch)
                        ,(generate1 ch defs) )
                      (look (,read-char ,port)) )) )
        ;;; Generate the look function in a file to help other ports.
      (call-with-output-file "look.scm"
        (lambda (out)
          (pp result out) ) )
      result ) ) )

;;; generate the looker that leave anything unchanged except instances of
;;; the following commands. This defines the function look.
;;; For those who cannot expand correctly this macro, the look function
;;; appears in the look.scm file.

;;; The run-XYZ functions takes the port as sole argument.

(generate-look-function
 ( ( "FromFile"            run-FromFile )
   ( "FromCFile"           run-FromCFile )
   ( "Eval"                run-Eval     )
   ( "ShowEvaluationOf"    run-ShowEvaluationOf )
   ( "PrettyPrint"         run-PrettyPrint )
   ( "PrettyFromFile"      run-PrettyFromFile )
   ( "Print"               run-Print    )
   ( "ExtractFromFile"     run-ExtractFromFile )
   ( "InsertFile"          run-InsertFile )
   ( "LoadFile"            run-LoadFile ) ) )

;;; end of find-string.scm
