;;; $Id: port-gambit.scm,v 1.5 2001/01/03 17:29:13 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file contains the functions that are highly dependent of the
;;; Scheme implementation and, here, Gambit from Marc Feeley.

;;; The name of the underlying Scheme system.

(define scm-name "Gambit")

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Utilities are mapcan, atom?, string-uppercase and
;;; string-lowercase. Some ports do not need to define them as they
;;; may already be present.

(define (mapcan fn l)
  (if (pair? l)
      (append (fn (car l))
              (mapcan fn (cdr l)) )
      '() ) )

(define (atom? e)
  (not (pair? e)) )

(define (string-uppercase string)
  (let* ((n (string-length string))
         (s (make-string n)) )
    (do ((i 0 (+ i 1)))
        ((= i n) s)
      (string-set! s i (char-upcase (string-ref string i))) ) ) )

(define (string-lowercase string)
  (let* ((n (string-length string))
         (s (make-string n)) )
    (do ((i 0 (+ i 1)))
        ((= i n) s)
      (string-set! s i (char-downcase (string-ref string i))) ) ) )

(define call/cc call-with-current-continuation)

(define (reverse! l)
  (define (nreverse l r)
    (if (pair? l)
        (let ((cdrl (cdr l)))
          (set-cdr! l r)
          (nreverse cdrl l) )
        r ) )
  (nreverse l '()) )

(define (symbol-append . args)
  (string->symbol
   (apply string-append
          (map (lambda (s)
                 (cond ((string? s) s)
                       ((symbol? s) (symbol->string s))
                       ((number? s) (number->string s))
                       (else (error 'symbol-append args)) ) )
               args ) ) ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Gambit dependencies.
;;; As of Gambit 2.7 (at least) gsi is case-sensitive.

(define system-can-peek-char? #t)

;;; The three POSIX streams.

(define stdin-port (current-input-port))
(define stdout-port (current-output-port))
(define stderr-port ##stderr)

;;; Defines case sensitivity.

(define *respect-input* #t)
(set! *respect-input* *respect-input*)
(define *uppercase-input* #f)
(set! *uppercase-input* *uppercase-input*)
(define *lowercase-input* #f)
(set! *lowercase-input* *lowercase-input*)

;;; Set how expressions to be evaluated (by Gambit) should be read.
;;; Gambit reads symbols in lowercase.

(define *evaluator-preferred-respect-input*   #t)
(set! *evaluator-preferred-respect-input* 
      *evaluator-preferred-respect-input* )
(define *evaluator-preferred-uppercase-input* #f)
(set! *evaluator-preferred-uppercase-input*
      *evaluator-preferred-uppercase-input* )
(define *evaluator-preferred-lowercase-input* #f)
(set! *evaluator-preferred-lowercase-input* 
      *evaluator-preferred-lowercase-input* )

;;; By default respect case sensitivity.

(define *respect-output* #t)
(set! *respect-output* *respect-output*)
(define *uppercase-output* #f)
(set! *uppercase-output* *uppercase-output*)
(define *lowercase-output* #f)
(set! *lowercase-output* *lowercase-output*)

;;; After each directive processing, flush buffers if not automatically
;;; provided by the implementation. 

(define (flush-all-buffers) 
  (flush-output stdout-port)
  (flush-output stderr-port) )

;;; probe-file tests if a file exists.

(define probe-file file-exists?)

;;; remove a file.

(define (delete-file filename)
  'nothing )

;;; Dynamically evaluates an expression. If an error occurs, wrap the
;;; error in a list with a distinguished symbol.

(define (LiSP2TeX-eval exp)
  (call/cc 
   (lambda (return)
     (##catch-all 
      (lambda (s args)
        ;; args contain cyclic values: don't try to print them!
        ;(define (simplify args)
        ;  (map ##decomp args) )
        (flush-all-buffers)
        (if *stop-at-first-error*
            (begin 
              ;; Don't print this, this is cyclic!
              (if (eq? s '##signal.global-unbound)
                  (let ((args ""))
                    (return (list '***EVALUATION-ERROR*** s args)) )
                  ;(##handle-interpreter-error
                  ; (##car args)
                  ; (##cadr args)
                  ; "Unbound variable:"
                  ; (##list (##decomp (##car args)))
                  ; '())
                  (##default-signal-catcher s args) )
              (exit 1) )
            (begin
              ;; Don't print this, this is cyclic!
              (if (eq? s '##signal.global-unbound)
                  (set! args "") )
              (return (list '***EVALUATION-ERROR*** s args)) ) ) )
      (lambda () (eval exp)) ) ) ) )

;;; exit prematurely

(define (LiSP2TeX-exit status)
  (flush-all-buffers)
  (exit status) )

;;; call unix

(define (LiSP2TeX-system cmd)
  (##shell-command cmd) )

;;; load the content of file.

(define (load-parameter-file file)
  ;;(load file)
  (let ((e* (file->list file)))
    (LiSP2TeX-eval `(begin #t . ,e*)) ) )

;; Read a whole file into a list of expressions.

(define (file->list filename)
  (call-with-input-file filename
    (lambda (in)
      (let loop ((e (read in)))
        (if (eof-object? e)
            '()
            (cons e (loop (read in))) ) ) ) ) )

;;; Define macros. They must be present for the compilation and the
;;; run-time since they are necessary to load customization files.
;;; New macros can be defined in customization files if they are defined
;;; with define-lisp2tex-macro.

(define-macro (define-lisp2tex-macro call . body)
  `(begin 
     (eval '(define-macro ,call . ,body))
     (define-macro ,call . ,body) ) )

;;; Make define-lisp2tex-macro pervasive itself  

(eval '(define-macro (define-lisp2tex-macro call . body)
         `(define-macro ,call . ,body) ))

;;; Gambit reads the entire file then expand each toplevel form in sequence.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Define some macros

;;; PC Scheme has still a binary apply so any port has to define this macro.
;;; But this macro is only needed for compilation.

(define-lisp2tex-macro (LiSP2TeX-apply . args)
  `(apply . ,args) )

(define-lisp2tex-macro (unless condition . body)
  `(if ,condition #f (begin . ,body)) )

(define-lisp2tex-macro (when condition . body)
  `(if ,condition (begin . ,body)) )

;;; end of port-gambit.scm
