;;; $Id: port-slib.scm,v 1.1 1994/09/29 14:17:37 queinnec Exp $
;;; Copyright (c) 1990-94 by Christian Queinnec. All rights reserved.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                        LiSP2TeX
;;;   Christian Queinnec             or to:  Christian Queinnec
;;;   <queinnec@polytechnique.fr>            <Christian.Queinnec@inria.fr>
;;;   Laboratoire d'Informatique de l'X      INRIA -- Rocquencourt
;;;   Ecole Polytechnique                    Domaine de Voluceau, BP 105
;;;   91128 Palaiseau                        78153 Le Chesnay Cedex
;;;   France                                 France
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; This program is distributed in the hope that it will be useful.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted, so long as the following
;;; conditions are met:
;;;      o credit to the authors is acknowledged following current
;;;        academic behaviour
;;;      o no fees or compensation are charged for use, copies, or
;;;        access to this software
;;;      o this copyright notice is included intact.
;;; This software is made available AS IS, and no warranty is made about
;;; the software or its performance.

;;; This file is based on the adaptation made by Fernando D. Mato Mira
;;; for Pseudo an adaptation of Scheme for Common Lisp.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Initialize Pseudo Scheme.

(ps:scheme)

;;; The name of the underlying Scheme system.

(define scm-name "pseudo")

;;; Missing stuff

(define (vector . args)
    (cl:make-array (cl:length args) :initial-contents args))

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Utilities are mapcan, atom?, string-uppercase and
;;; string-lowercase. Some ports do not need to define them as they
;;; may already be present.

(define mapcan #'cl:mapcan)

(define atom? #'cl:atom)

(define string-uppercase #'cl:string-upcase)

(define string-lowercase #'cl:string-downcase)

(define reverse! #'cl:nreverse)

(define (symbol-append . args)
  (string->symbol
   (apply string-append
          (map (lambda (s)
                 (cond ((string? s) s)
                       ((symbol? s) (symbol->string s))
                       ((number? s) (number->string s))
                       (else (error 'symbol-append args)) ) )
               args ) ) ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; Only used in their dynamic extent so bind-exit is safe.

(define call/cc call-with-current-continuation)

;;; pseudoscheme has a working peek-char.

(define system-can-peek-char? #t)

;;; The three POSIX streams.

(define stdin-port  cl:*standard-input*)
(define stdout-port cl:*standard-output*)
(define stderr-port cl:*error-output*)

;;; Defines case sensitivity.

(define *respect-input* #t)
(define *uppercase-input* #f)
(define *lowercase-input* #f)

;;; Set how expressions to be evaluated (by pseudoscheme) should be read.
;;; Pseudoscheme reads symbols in uppercase.

(define *evaluator-preferred-respect-input*   #f)
(define *evaluator-preferred-uppercase-input* #t)
(define *evaluator-preferred-lowercase-input* #f)

;;; By default respect case sensitivity.

(define *respect-output* #t)
(define *uppercase-output* #f)
(define *lowercase-output* #f)

;;; After each directive processing, flush buffers if not automatically
;;; provided by the implementation. 

(define (flush-all-buffers) 
  (cl:force-output stdout-port)
  (cl:force-output stderr-port ) )

;;; probe-file tests if a file exists.

(define probe-file #'cl:probe-file)

;;; Evaluate an expression and catch any error that may occur.
;;; The ps:lisp2tex-eval function is defined in lisp2tex.lisp since
;;; it uses a CL macro and must be expanded with CL.

(define LiSP2TeX-eval #'ps:lisp2tex-eval)

;;; load the content of file. Load it non verbosely to not clutter the
;;; output stream.  The content of the file may use the various macros
;;; that are defined in emit.scm. It is better to load the file not in
;;; verbose mode to avoid cluttering the output: loadq does this.

(define load-parameter-file #'load)

;;; Define macros. They must be present for the compilation and the
;;; run-time since they are necessary to load customization files.
;;; New macros can be defined in customization files if they are defined
;;; with define-lisp2tex-macro.

;(define-syntax define-lisp2tex-macro
;  (syntax-rules ()
;    ((_ (name . variables) . body)
;     (define-syntax name
;       (syntax-rules ()
;         ((__ . variables) (begin . body)) ) ) ) ) )

(define-syntax define-lisp2tex-macro
  (lambda (e r c)
    (let* ((call      (cadr e))
           (name      (car call))
           (variables (cdr call))
           (body      (cddr e)) )
      `(define-syntax ,name
         (lambda (e1 r1 c1)
           (apply (lambda ,variables . ,body) (cdr e1)) ) ) ) ) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Define some macros. The three following forms should be correct if
;;; the previous define-lisp2tex-macro is. Since I do not have access
;;; to Pseudo to test that, I give the expanded form.

;;; PC Scheme has still a binary apply so any port has to define this macro.
;;; But this macro is only needed for compilation.

;(define-lisp2tex-macro (LiSP2TeX-apply . args)
;  `(apply . ,args) )

;(define-lisp2tex-macro (unless condition . body)
;  `(if ,condition #f (begin . ,body)) )

;(define-lisp2tex-macro (when condition . body)
;  `(if ,condition (begin . ,body)) )

;;; These are equivalent to the direct form:

(define-syntax Lisp2TeX-apply
  (syntax-rules ()
    ((LiSP2TeX-apply args ...)
     (apply args ...) ) ) )

(define-syntax unless
  (syntax-rules ()
    ((unless condition form ...)
     (if (not condition) (begin form ...)) ) ) )

(define-syntax when
  (syntax-rules ()
    ((when condition form ...)
     (if condition (begin form ...)) ) ) )

;;;==================================================================
;;; this is the end of the normal port-slib.scm file.
;;; Now load the rest of the common source files of LiSP2TeX.

;;;(load "read-objects.scm")
;;;(load "read.scm")
;;;(load "scan-objects.scm")
;;;(load "scan-file.scm")
;;;(load "emit.scm")
;;;(load "pp.scm")
;;;(load "commands.scm")
;;;(load "version.scm")
;;;(load "look.scm")

;;; end of port-pseudo.scm
