;;; $Id: port-slib.scm,v 1.2 1998/08/26 18:40:15 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file loads LiSP2TeX under scm and slib. It is patterned after the
;;; other port-*.scm files of the current directory.

;;; The name of the underlying Scheme system.

(define scm-name "scm+slib")

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Utilities are mapcan, atom?, string-uppercase and
;;; string-lowercase. Some ports do not need to define them as they
;;; may already be present.

(define (mapcan fn l)
  (if (pair? l)
      (append (fn (car l))
              (mapcan fn (cdr l)) )
      '() ) )

(define (atom? e)
  (not (pair? e)) )

(define (string-uppercase string)
  (let* ((n (string-length string))
         (s (make-string n)) )
    (do ((i 0 (+ i 1)))
        ((= i n) s)
      (string-set! s i (char-upcase (string-ref string i))) ) ) )

(define (string-lowercase string)
  (let* ((n (string-length string))
         (s (make-string n)) )
    (do ((i 0 (+ i 1)))
        ((= i n) s)
      (string-set! s i (char-downcase (string-ref string i))) ) ) )

(define (reverse! l)
  (define (nreverse l r)
    (if (pair? l)
       (let ((cdrl (cdr l)))
         (set-cdr! l r)
         (nreverse cdrl l) )
       r ) )
  (nreverse l '()) )

(define (symbol-append . args)
  (string->symbol
   (apply string-append
          (map (lambda (s)
                 (cond ((string? s) s)
                       ((symbol? s) (symbol->string s))
                       ((number? s) (number->string s))
                       (else (error 'symbol-append args)) ) )
               args ) ) ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; Only used in their dynamic extent so bind-exit is safe.

(define call/cc call-with-current-continuation)

;;; scm has a working peek-char.

(define system-can-peek-char? #t)

;;; The three POSIX streams.

(define stdin-port (current-input-port))
(define stdout-port (current-output-port))
(define stderr-port (current-error-port))

;;; Defines case sensitivity.

(define *respect-input* #t)
(define *uppercase-input* #f)
(define *lowercase-input* #f)

;;; Set how expressions to be evaluated (by Bigloo) should be read.
;;; Bigloo reads symbols in uppercase.

(define *evaluator-preferred-respect-input*   #f)
(define *evaluator-preferred-uppercase-input* #f)
(define *evaluator-preferred-lowercase-input* #t)

;;; By default respect case sensitivity.

(define *respect-output* #t)
(define *uppercase-output* #f)
(define *lowercase-output* #f)

;;; After each directive processing, flush buffers if not automatically
;;; provided by the implementation. 

(define (flush-all-buffers) 
  (force-output stdout-port)
  (force-output stderr-port ) )

;;; probe-file tests if a file exists.

(define probe-file file-exists?)

;;; This should catch any errors even if detected by the underlying Scheme
;;; system. I do not know how to do it in scm.            TO BE DONE...
;;; This means that LiSP2TeX probably aborts on the first error it finds.

(define (LiSP2TeX-eval exp)
  (eval exp) )

;;; load the content of file. Load it non verbosely to not clutter the
;;; output stream.  The content of the file may use the various macros
;;; that are defined in emit.scm. It is better to load the file not in
;;; verbose mode to avoid cluttering the output: loadq does this.

(define (load-parameter-file file)
  (load file) )

;;; Define macros. They must be present for the compilation and the
;;; run-time since they are necessary to load customization files.
;;; New macros can be defined in customization files if they are defined
;;; with define-lisp2tex-macro.

(require 'defmacro)

(defmacro define-lisp2tex-macro (call . body)
  `(defmacro ,(car call) ,(cdr call) . ,body) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Define some macros

;;; PC Scheme has still a binary apply so any port has to define this macro.
;;; But this macro is only needed for compilation.

(define-lisp2tex-macro (LiSP2TeX-apply . args)
  `(apply . ,args) )

(define-lisp2tex-macro (unless condition . body)
  `(if ,condition #f (begin . ,body)) )

(define-lisp2tex-macro (when condition . body)
  `(if ,condition (begin . ,body)) )

;;; this is the end of the normal port-slib.scm file.
;;; Now load the rest of the common source files of LiSP2TeX.

(load "read-objects.scm")
(load "read.scm")
(load "scan-objects.scm")
(load "scan-file.scm")
(load "emit.scm")
(load "pp.scm")
(load "commands.scm")
(load "version.scm")
(load "look.scm")

;;; and start the filtering process...
;;; Since scm is called with a lot of arguments, I took the convention
;;; that options start after the -- option. This is consistant with scm.

(let ((args (cdr (member "--" *argv*))))
  (LiSP2TeX args)  )

;;; end of slib.scm
