;;; $Id: read-objects.scm,v 1.2 1998/08/26 18:40:16 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file defines how symbols and strings are created by the reader.
;;; An extensible string is used to incrementally build symbols or strings.
;;; It is at the end finalized into the appropriate data structure.

;;; add a character to a symbol under construction.
;;; The boolean `respect' if true imposes that the character will not be
;;; turned to upper or lower case.
(define (increment-symbol char sm respect)
  (increment-string 
   (cond (respect char)
         (*respect-input* char)
         (*uppercase-input* (char-upcase char))
         (*lowercase-input* (char-downcase char))
         (else char) )
   sm ) )

;;; add a character to an extensible string
(define (increment-string char sm)
  (let ((string (extensible-string-characters sm))
        (index (extensible-string-index sm)) )
    (string-set! string index char)
    (set-extensible-string-index! sm (+ 1 index))
    (if (< (+ 1 index) (string-length string))
        sm
        (let* ((newindex (+ 1 index))
               (newstring (make-string (* 2 newindex))) )
          (set-extensible-string-characters! sm newstring)
          (do ((i 0 (+ 1 i)))
              ((= i newindex) sm)
            (string-set! newstring i (string-ref string i)) ) ) ) ) )

;;; takes an extensible string and returns a symbol
(define (finalize-symbol sm)
  (string->symbol (finalize-string sm)) )

;;; takes an extensible string and returns a string
(define (finalize-string sm)
  (substring (extensible-string-characters sm)
             0 
             (extensible-string-index sm) ) )


;;; An extensible string is extended as needed.
(define (make-extensible-string initial-size)
  (vector 'extensible-string (make-string initial-size) 0) )
(define (extensible-string-characters sm)
  (vector-ref sm 1) )
(define (set-extensible-string-characters! sm newstring)
  (vector-set! sm 1 newstring) )
(define (extensible-string-index sm)
  (vector-ref sm 2) )
(define (set-extensible-string-index! sm index)
  (vector-set! sm 2 index) )

;;; end of read-object.scm
