;;; $Id: scan-c.scm,v 1.4 1996/03/31 12:52:17 queinnec Exp $
;;; Contribution from Thierry Saura.
;;; This file defines how to scan a .c or .h file.
;;; The main function is scan-c-file which returns a scanned-c-file object.
;;; This file also uses the seconds.scm file.

;;; This structure is similar to scanned-file except that it has a 
;;; different tag. 

(define (make-scanned-c-file filename scanned-c-expressions string cinfos)
  (vector scan-c-file
          filename scanned-c-expressions string cinfos ) )
(define (scanned-c-file-filename sf)
  (vector-ref sf 1) )
(define (scanned-c-file-scanned-expressions sf)
  (vector-ref sf 2) )
(define (scanned-c-file-string sf)
  (vector-ref sf 3) )
(define (scanned-c-file-cinfos sf)
  (vector-ref sf 4) )

;;; This structure defines a C expression. The internal comments are held
;;; in a list of (start . end) pairs (where start is the first character to 
;;; skip and end the first one not to skip).
;;;
;;; Between start and body is the header
;;; of a C definition and between body and end its body. 
;;; start designates the first meaningful character,
;;; body designates the { starting the body,
;;; end designates the character that follows the ; that ends the definition.

(define (make-scanned-c-expression start end bodystart bodyend comments)
  (vector 'scanned-c-expression start end bodystart bodyend comments) )
(define (scanned-c-expression-start se)
  (vector-ref se 1) )
(define (scanned-c-expression-end se)
  (vector-ref se 2) )
(define (scanned-c-expression-bodystart se)
  (vector-ref se 3) )
(define (scanned-c-expression-bodyend se)
  (vector-ref se 4) )
(define (scanned-c-expression-comments se)
  (vector-ref se 5) )

;;;==================================================================
; {{{ File Description
; Creation : Jan 11, 1996
; We define a function which parses a c file to find function's definitions
; }}}

; {{{ Structures definition
; c-info = <info begin end>
; info = <fdef> | comments | string | ender | first

(define (c-info info)
  (vector 'c-info info 0 0))

(define (make-c-info info beg end)
  (vector 'c-info info beg end))

(define (c-info? obj)
  (and (vector? obj)
       (eq? 'c-info (vector-ref obj 0))))

(define (c-info-info obj)
  (vector-ref obj 1))

(define (c-info-info-set! obj val)
  (vector-set! obj 1 val)
  obj )

(define (c-info-begin obj)
  (vector-ref obj 2))

(define (c-info-begin-set! obj val)
  (vector-set! obj 2 val)
  obj )

(define (c-info-end obj)
  (vector-ref obj 3))

(define (c-info-end-set! obj val)
  (vector-set! obj 3 val)
  obj )

(define (fdef? a-info)
  (and (integer? (c-info-info a-info)) (<= 0 (c-info-info a-info))))

(define (comments? a-info)
  (eq? 'comments (c-info-info a-info)))

(define (astring? a-info)
  (eq? 'string (c-info-info a-info)))

(define (ender? a-info)
  (eq? 'ender (c-info-info a-info)))

(define (macro? a-info)
  (eq? 'macro (c-info-info a-info)))

(define (first? a-info)
  (eq? 'first (c-info-info a-info)))

(define (init? a-info)
  (eq? 'init (c-info-info a-info)))

(define (vardef? a-info)
  (eq? 'vardef (c-info-info a-info)))

;;; Useless and clashes with Bigloo's own struct?
;(define (struct? a-info)
;  (eq? 'struct (c-info-info a-info)))

(define (typedef? a-info)
  (eq? 'typedef (c-info-info a-info)))
; }}}

; {{{ Counting-streams
(define (end-of-stream? obj)
  (eof-object? obj))

(define (counting-stream-extract stream)
  (counting-stream-string stream))
; }}}

; {{{ State structure
;(define-structure state (level 0) (c-info '()))

(define (state)
  (vector 'state 0 '()))

(define (make-state level list)
  (vector 'state level list))

(define (state? obj)
  (and (vector? obj)
       (eq? 'state (vector-ref obj 0))))

(define (state-level obj)
  (vector-ref obj 1))

(define (state-level-set! obj val)
  (vector-set! obj 1 val)
  obj )

(define (state-c-info obj)
  (vector-ref obj 2))

(define (state-c-info-set! obj val)
  (vector-set! obj 2 val)
  obj )

(define (push-c-info the_state an_info)
  (state-c-info-set! the_state (cons an_info (state-c-info the_state))))

(define (pop-c-info the_state)
  (state-c-info-set! the_state (cdr (state-c-info the_state))))

(define (level++ the_state)
  (state-level-set! the_state (+ 1 (state-level the_state))))

(define (level-- the_state)
  (state-level-set! the_state (- (state-level the_state) 1)))
; }}}

; {{{ Useful functions
(define (last-first-position the_state)
  (let loop ((l (state-c-info the_state)))
    (if (pair? l)
	(let ((element (car l)))
	  (if (first? element)
	      (c-info-begin element)
	      (loop (cdr l))))
	0)))

(define (end-of-function-set! the_state pos)
  (let loop ((l (state-c-info the_state)))
    (if (pair? l)
	(let ((element (car l)))
	  (if (fdef? element)
	      (begin
		(c-info-end-set! element pos)
		the_state)
	      (loop (cdr l))))
	the_state)))

(define (end-of-init-set! the_state pos)
  (let loop ((l (state-c-info the_state)))
    (if (pair? l)
	(let ((element (car l)))
	  (if (init? element)
	      (begin
		(c-info-end-set! element pos)
		the_state)
	      (loop (cdr l))))
	the_state)))

(define (end-of-macro-set! the_state pos)
  (let loop ((l (state-c-info the_state)))
    (if (pair? l)
	(let ((element (car l)))
	  (if (macro? element)
	      (begin
		(c-info-end-set! element pos)
		the_state)
	      (loop (cdr l))))
	the_state)))

(define (set-c-info-end! the_state pos)
  (c-info-end-set! (car (state-c-info the_state)) pos))
; }}} 

; -------------------------------------------------------------------
; This is the main state
(define (parse-c-aux stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(begin
	  (cond
	   ((char=? obj #\{)
	    (parse-function stream
			    (push-c-info the_state
					 (make-c-info 
					  (last-first-position the_state)
					  (counting-stream-position stream)
					  0))))
	   ((char=? obj #\/)
	    (parse-begin-comments
	     stream
	     (push-c-info the_state
			  (make-c-info
			   'comments
			   (counting-stream-position stream)
			   0))))
	   ((char=? obj #\=)
	    (parse-initialisation stream
				  (push-c-info
				   (push-c-info
				    the_state
				    (make-c-info
				     'vardef
				     (last-first-position the_state)
				     (- (counting-stream-position
					 stream) 1)))
				   (make-c-info
				    'init
				    (counting-stream-position stream)
				    0))))
	   ((char=? obj #\")
	    (parse-string
	     stream
	     (push-c-info
	      the_state
	      (make-c-info 'string
			   (counting-stream-position stream)
			   0))))
	   ((char=? obj #\;)
	    (push-c-info the_state
			 (let ((pos (counting-stream-position stream)))
			   (make-c-info 'ender pos pos))))
	   ((char=? obj #\#)
	    (parse-begin-macro stream
			       (push-c-info the_state
					    (make-c-info
					     'macro
					     (counting-stream-position stream)
					     0))))
	   ((or (char=? obj #\newline)
		(char=? obj #\tab)
		(char=? obj #\space))
	    the_state)
	   (else (parse-first stream
			      (push-c-info the_state
					   (let ((pos (counting-stream-position
						       stream)))
					     (make-c-info
					      'first
					      pos
					      pos))))))
	  (loop (counting-stream-read-char stream))))))

; -------------------------------------------------------------------
; In this state, we suppose to have find the first interesting
; char, we verify if it's true
(define (parse-first stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\{)
	  (parse-function stream
			  (push-c-info the_state
				       (make-c-info 
					(last-first-position the_state)
					(counting-stream-position stream)
					0))))
	 ((char=? obj #\=)
	  (parse-initialisation stream
				(push-c-info
				 (push-c-info
				  the_state
				  (make-c-info
				   'vardef
				   (last-first-position the_state)
				   (- (counting-stream-position
				       stream) 1)))
				 (make-c-info
				  'init
				  (counting-stream-position stream)
				  0))))
	 ((char=? obj #\/)
	  (parse-begin-comments
	   stream
	   (push-c-info the_state
			(make-c-info
			 'comments
			 (counting-stream-position stream)
			 0)))
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\")
	  (parse-string
	   stream
	   (push-c-info
	    the_state
	    (make-c-info 'string
			 (counting-stream-position stream)
			 0)))
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\#)
	  (parse-begin-macro stream
			     (push-c-info the_state
					  (make-c-info
					   'macro
					   (counting-stream-position stream)
					   0)))
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\;)
	  (push-c-info the_state
		       (let ((pos (counting-stream-position stream)))
			 (make-c-info 'ender pos pos))))
	 (else (loop (counting-stream-read-char stream)))))))

; --------------------------------------------------------------------
; We are in a function
(define (parse-function stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\{)
	  (level++ the_state)
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\})
	  (if (= 0 (state-level the_state))
	      (end-of-function-set! the_state
				   (counting-stream-position stream))
	      (begin
		(level-- the_state)
		(loop (counting-stream-read-char stream)))))
	 ((char=? obj #\/)
	  (parse-begin-comments
	   stream
	   (push-c-info the_state
			(make-c-info
			 'comments
			 (counting-stream-position stream)
			 0)))
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\")
	  (parse-string stream
			(push-c-info the_state
				     (make-c-info
				      'string
				      (counting-stream-position stream)
				      0)))
	  (loop (counting-stream-read-char stream)))
	 (else (loop (counting-stream-read-char stream)))))))

; --------------------------------------------------------------------
; We parse an initialisation
(define (parse-initialisation stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\{)
	  (level++ the_state)
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\})
	  (level-- the_state)
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\/)
	  (parse-begin-comments
	   stream
	   (push-c-info the_state
			(make-c-info
			 'comments
			 (counting-stream-position stream)
			 0)))
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\")
	  (parse-string
	   stream
	   (push-c-info
	    the_state
	    (make-c-info 'string
			 (counting-stream-position stream)
			 0)))
	  (loop (counting-stream-read-char stream)))
	 ((and (char=? obj #\;)
	       (= 0 (state-level the_state)))
	  (push-c-info
	   (end-of-init-set! the_state (counting-stream-position stream))
	   (let ((pos (counting-stream-position stream)))
			   (make-c-info 'ender pos pos))))
	 (else (loop (counting-stream-read-char stream)))))))

; --------------------------------------------------------------------
; It's perhaps the begin of a comment
(define (parse-begin-comments stream the_state)
  (let ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\*)
	  (parse-comments stream the_state))
	 ((char=? obj #\/)
	  (parse-comments-c++ stream the_state))
	 (else (pop-c-info the_state))))))

; --------------------------------------------------------------------
; We are in a comment
(define (parse-comments stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\*)
	  (parse-end-comments stream the_state))
	 (else (loop (counting-stream-read-char stream)))))))

; --------------------------------------------------------------------
; It's perhaps the end of a comment
(define (parse-end-comments stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\*)
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\/)
	  (set-c-info-end! the_state (counting-stream-position stream)))
	 (else (parse-comments stream the_state))))))

; --------------------------------------------------------------------
; We are in a C++ comment
(define (parse-comments-c++ stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\newline)
	  (set-c-info-end! the_state (counting-stream-position stream)))
	 (else (loop (counting-stream-read-char stream)))))))

; --------------------------------------------------------------------
; We are in a string
(define (parse-string stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\")
	  (set-c-info-end! the_state (counting-stream-position stream)))
	 (else (loop (counting-stream-read-char stream)))))))

; --------------------------------------------------------------------
; It's perhaps the begin of a macro
(define (parse-begin-macro stream the_state)
  (let ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\#)
	  (pop-c-info the_state))
	 (else (parse-macro stream the_state))))))

; --------------------------------------------------------------------
; We are in a macro
(define (parse-macro stream the_state)
  (let loop ((obj (counting-stream-read-char stream)))
    (if (end-of-stream? obj)
	the_state
	(cond
	 ((char=? obj #\\)
	  (level++ the_state)
	  (loop (counting-stream-read-char stream)))
	 ((char=? obj #\newline)
	  (if (= 0 (state-level the_state))
	      (end-of-macro-set! the_state (counting-stream-position stream))
	      (begin (level-- the_state)
		     (loop (counting-stream-read-char stream)))))
	 ((char=? obj #\")
	  (parse-string stream
			(push-c-info
			 the_state
			 (make-c-info 'string
				      (counting-stream-position stream)
				      0)))
	  (loop (counting-stream-read-char stream)))
	 (else (loop (counting-stream-read-char stream)))))))

; {{{ Starting function
;(define (scan-c-file name)
;  (call-with-input-file name
;    (lambda (port)
;      (let loop ((res (parse-c (counting-stream port)
;			       (state))))
;	(if (pair? res)
;	    (let ((an-info (car res)))
;	      (cond
;	       ((fdef? an-info)
;		(print "Accolades de premier niveau")
;		(print "     Debut du nom : " (c-info-info an-info))
;		(print "     Debut : " (c-info-begin an-info))
;		(print "     Fin : " (c-info-end an-info))
;		(loop (cdr res)))
;	       ((comments? an-info)
;		(print "Commentaire")
;		(print "     Debut : " (c-info-begin an-info))
;		(print "     Fin : " (c-info-end an-info))
;		(loop (cdr res)))
;	       ((astring? an-info)
;		(print "Chaine")
;		(print "     Debut : " (c-info-begin an-info))
;		(print "     Fin : " (c-info-end an-info))
;		(loop (cdr res)))
;	       ((ender? an-info)
;		(print "Terminateur")
;		(print "     Position : " (c-info-end an-info))
;		(loop (cdr res)))
;	       ((macro? an-info)
;		(print "Macro")
;		(print "     Debut : " (c-info-begin an-info))
;		(print "     Fin : " (c-info-end an-info))
;		(loop (cdr res)))
;	       ((first? an-info)
;		(print "Premier caractere interessant")
;		(print "     Position : " (c-info-end an-info))
;		(loop (cdr res)))
;	       (else (error 'scan-c-file "Unknown type"
;			    (c-info-info an-info)))))
;	    #t))))) 
; }}}

;(define (parse-c stream the_state)
;  (reverse (state-c-info (parse-c-aux stream the_state))))

;(define (start args)
;  (let ((arg (cdr args)))
;    (if (pair? arg)
;	(scan-c-file (car arg))
;	(print "syntax : nparse-c <file>"))))

; {{{ Starting function

(define (scan-c-file filename)
  (call-with-input-file filename
    (lambda (in)
      (let ((extensible-string (make-extensible-string 4096))
	    (eof-object 'wait) )
	;; read the whole file into an extensible-string.
	;; This code is completely inefficient. It should be better
	;; to determine the size of the file and to read it in one go.
	(letrec ((loop 
		  (lambda (char)
		    (cond 
		     ((eof-object? char)
		      (set! eof-object char) )
		     (else 
		      (increment-string char extensible-string)
		      (loop (read-char in)) ) ) ) ))
	  (loop (read-char in)) )
	;; determine the bounds of each Sexpression
	(let* ((string (finalize-string extensible-string))
               (cs (parse-c (make-counting-stream
                             filename
                             string
                             (extensible-string-index extensible-string)
                             0
                             eof-object
                             0 )
                            (state) )) )
          ;;(display cs)(newline) ; DEBUG
	  ;; build the final object
	  (make-scanned-c-file
           filename
           (apply vector cs)
           string
           cs ) ) ) ) ) )

(define (parse-c stream the_state)
  (Second_Scan stream (state-c-info (parse-c-aux stream the_state))))

; }}}

;;; end of scan-c.scm
